
PORTAL_FILE_SCHEMA = {
    "type": "object",
    "properties": {
        "resourceType": {"type": "string"},
        "apiVersion": {"type": "string"},
        "learnMore": {
            "type": "object",
            "properties": {
                "url": {"type": "string"}
            }
        },
        "commands": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                    "description": {"type": "string"},
                    "path": {"type": "string"},
                    "confirmation": {"type": "boolean"},
                    "help": {
                        "type": "object",
                        "properties": {
                            "learnMore": {
                                "type": "object",
                                "properties": {
                                    "url": {"type": "string"}
                                }
                            },
                            "parameterSets": {
                                "type": "array",
                                "items": {
                                    "type": "object",
                                    "properties": {
                                        "parameters": {
                                            "type": "array",
                                            "items": {"type": "string"}
                                        }
                                    }
                                }
                            }
                        }
                    },
                    "examples": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "properties": {
                                "description": {"type": "string"},
                                "parameters": {
                                    "type": "array",
                                    "items": {
                                        "type": "object",
                                        "properties": {
                                            "name": {"type": "string"},
                                            "value": {"type": "string"}
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}



from schematics.models import Model
from schematics.types import StringType
from schematics.types.serializable import serializable

from ._fields import CMDVariantField, CMDBooleanField


class CMDOutput(Model):
    # properties as tags
    TYPE_VALUE = None  # types: "array", "object", "string",

    class Options:
        serialize_when_none = False

    @serializable
    def type(self):
        return self._get_type()

    def _get_type(self):
        assert self.TYPE_VALUE is not None
        return self.TYPE_VALUE

    @classmethod
    def _claim_polymorphic(cls, data):
        if cls.TYPE_VALUE is None:
            return False

        if isinstance(data, dict):
            type_value = data.get('type', None)
            if type_value is not None:
                typ = type_value.replace("<", " ").replace(">", " ").strip().split()[0]
                return typ == cls.TYPE_VALUE
        elif isinstance(data, CMDOutput):
            return data.TYPE_VALUE == cls.TYPE_VALUE
        return False


class CMDObjectOutput(CMDOutput):
    TYPE_VALUE = 'object'

    ref = CMDVariantField(required=True)

    client_flatten = CMDBooleanField(
        serialized_name='clientFlatten',
        deserialize_from='clientFlatten',
    )


class CMDArrayOutput(CMDOutput):
    TYPE_VALUE = 'array'

    ref = CMDVariantField(required=True)

    client_flatten = CMDBooleanField(
        serialized_name='clientFlatten',
        deserialize_from='clientFlatten',
    )
    next_link = CMDVariantField(
        serialized_name='nextLink',
        deserialize_from='nextLink',
    )


class CMDStringOutput(CMDOutput):
    TYPE_VALUE = 'string'

    # ref means get value from content
    ref = CMDVariantField()
    # use placeholder value as output
    value = StringType()



from schematics.models import Model
from schematics.types import ModelType

from ._content import CMDRequestJson


class CMDHttpRequestBody(Model):
    POLYMORPHIC_KEY = None

    @classmethod
    def _claim_polymorphic(cls, data):
        if cls.POLYMORPHIC_KEY is None:
            return False

        if isinstance(data, dict):
            return cls.POLYMORPHIC_KEY in data
        elif isinstance(data, CMDHttpRequestBody):
            return hasattr(data, cls.POLYMORPHIC_KEY)

        return False

    def generate_args(self, ref_args):
        raise NotImplementedError()

    def diff(self, old, level):
        raise NotImplementedError()

    def reformat(self, **kwargs):
        return NotImplementedError()

    def register_cls(self, **kwargs):
        raise NotImplementedError()


class CMDHttpRequestJsonBody(CMDHttpRequestBody):
    POLYMORPHIC_KEY = "json"

    json = ModelType(CMDRequestJson, required=True)

    def generate_args(self, ref_args):
        return self.json.generate_args(ref_args=ref_args)

    def diff(self, old, level):
        if not isinstance(old, self.__class__):
            return f"Response type changed: '{type(old)}' != '{self.__class__}'"
        return self.json.diff(old.json, level)

    def reformat(self, **kwargs):
        self.json.reformat(**kwargs)

    def register_cls(self, **kwargs):
        self.json.register_cls(**kwargs)




from flask import Blueprint
import click, logging
from utils.config import Config
from cli.controller.portal_cli_generator import PortalCliGenerator

logging.basicConfig(level="INFO")

bp = Blueprint('portal', __name__, url_prefix='/CLI/Portal')

# commands
@bp.cli.command("generate", short_help="Generate command portal json file.")
@click.option(
    "--aaz-path", '-a',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.AAZ_PATH,
    required=not Config.AAZ_PATH,
    callback=Config.validate_and_setup_aaz_path,
    expose_value=False,
    help="The local path of aaz repo."
)
@click.option(
    "--cli-path", '-c',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.CLI_PATH,
    required=not Config.CLI_PATH,
    callback=Config.validate_and_setup_cli_path,
    expose_value=False,
    help="The local path of azure-cli repo. Official repo is https://github.com/Azure/azure-cli"
)
@click.option(
    "--cli-extension-path", '-e',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.CLI_EXTENSION_PATH,
    required=not Config.CLI_EXTENSION_PATH,
    callback=Config.validate_and_setup_cli_extension_path,
    expose_value=False,
    help="The local path of azure-cli-extension repo. Official repo is https://github.com/Azure/azure-cli-extensions"
)
@click.option(
    "--module", '-m',
    required=True, multiple=True,
    help="Name of the module to generate seperated."
)
def generate_module_command_portal(module):
    from cli.controller.az_module_manager import AzMainManager, AzExtensionManager
    from command.controller.specs_manager import AAZSpecsManager
    az_main_manager = AzMainManager()
    az_ext_manager = AzExtensionManager()
    aaz_spec_manager = AAZSpecsManager()
    root = aaz_spec_manager.find_command_group()
    if not root:
        return "Command group spec root not exist"
    portal_cli_generator = PortalCliGenerator()
    for mod in module:
        if az_main_manager.has_module(mod):
            cli_module = az_main_manager.load_module(mod)
            registered_cmds = az_main_manager.find_module_cmd_registered(cli_module['profiles']['latest'])
            logging.info("Input module: {0} in cli".format(mod))
        elif az_ext_manager.has_module(mod):
            cli_module = az_ext_manager.load_module(mod)
            registered_cmds = az_ext_manager.find_module_cmd_registered(cli_module['profiles']['latest'])
            logging.info("Input module: {0} in cli extension".format(mod))
        else:
            raise ValueError("Invalid input module: {0}, please check".format(mod))
        cmd_portal_list = portal_cli_generator.generate_cmds_portal_info(aaz_spec_manager, registered_cmds)
        portal_cli_generator.generate_cmds_portal_file(cmd_portal_list)
    logging.info("done")



from schematics.models import Model
from schematics.types import StringType, ModelType, BooleanType, ListType
from .parameter import ParameterField, ParameterBase
from .reference import Linkable, Reference


class XmsParameterizedHost(Model, Linkable):
    """
    When used, replaces the standard OpenAPI "host" attribute with a host that contains variables to be replaced as part of method execution or client construction, very similar to how path parameters work.

    https://github.com/Azure/autorest/tree/main/docs/extensions#x-ms-parameterized-host
    """

    host_template = StringType(
        required=True,
        serialized_name="hostTemplate",
        deserialize_from="hostTemplate"
    )  # Specifies the parameterized template for the host.
    use_scheme_prefix = BooleanType(
        default=True,
        serialized_name="useSchemePrefix",
        deserialize_from="useSchemePrefix"
    )  # Specifies whether to prepend the default scheme a.k.a protocol to the base uri of client.
    position_in_operation = StringType(
        default="first",
        choices=("first", "last"),
        serialized_name="positionInOperation",
        deserialize_from="positionInOperation"
    )  # Specifies whether the list of parameters will appear in the beginning or in the end, in the method signature for every operation. The order within the parameters provided in the below mentioned array will be preserved. Either the array of parameters will be prepended or appended, based on the value provided over here. Valid values are "first", "last". Every method/operation in any programming language has parameters categorized into two buckets "required" and "optional". It is natural for optional parameters to appear in the end in a method signature. This aspect will be preserved, while prepending(first) or appending(last) hostTemplate parameters.
    parameters = ListType(ParameterField(support_reference=True))  # The list of parameters that are used within the hostTemplate. This can include both reference parameters as well as explicit parameters. Note that "in" is required and must be set to "path". The reference parameters will be treated as global parameters and will end up as property of the client.

    def link(self, swagger_loader, *traces):
        if self.is_linked():
            return
        super().link(swagger_loader, *traces)

        if self.parameters is not None:
            for idx, param in enumerate(self.parameters):
                if isinstance(param, Linkable):
                    param.link(swagger_loader, *self.traces, 'parameters', idx)

            # replace parameter reference by parameter instance
            for idx in range(len(self.parameters)):
                param = self.parameters[idx]
                while isinstance(param, Reference):
                    param = param.ref_instance
                assert isinstance(param, ParameterBase)
                self.parameters[idx] = param


class XmsParameterizedHostField(ModelType):

    def __init__(self, **kwargs):
        super(XmsParameterizedHostField, self).__init__(
            XmsParameterizedHost,
            serialized_name="x-ms-parameterized-host",
            deserialize_from="x-ms-parameterized-host",
            **kwargs
        )



import os
import shutil
from cli.templates import get_templates
from command.model.configuration import CMDCommand
from utils.case import to_snake_case
from .az_command_generator import AzCommandGenerator
from utils import exceptions


class AzProfileGenerator:
    """Used to generate atomic layer command group"""

    def __init__(self, aaz_folder, profile):
        self.aaz_folder = aaz_folder
        self.profile = profile
        self.profile_folder_name = profile.profile_folder_name
        self._removed_folders = set()
        self._removed_files = set()
        self._modified_files = {}

    def generate(self):
        # check aaz/__init__.py
        file_name = '__init__.py'
        if not self._exist_file(file_name):
            tmpl = get_templates()['aaz'][file_name]
            data = tmpl.render()
            self._update_file(file_name, data=data)

        if not self.profile.command_groups:
            # remove the whole aaz/{profile}
            self._delete_folder(self.profile_folder_name)
        else:
            # check aaz/{profile}/__init__.py
            file_name = '__init__.py'
            if not self._exist_file(self.profile_folder_name, file_name):
                tmpl = get_templates()['aaz']['profile'][file_name]
                data = tmpl.render()
                self._update_file(self.profile_folder_name, file_name, data=data)

            remain_folders, _ = self._list_package(self.profile_folder_name)
            for command_group in self.profile.command_groups.values():
                assert len(command_group.names) == 1, f"Invalid command group name: {command_group.names}"
                command_group_folder_name = command_group.names[-1].replace('-', '_')
                self._generate_by_command_group(
                    profile_folder_name=self.profile_folder_name,
                    command_group=command_group
                )
                if command_group_folder_name in remain_folders:
                    remain_folders.remove(command_group_folder_name)
            for name in remain_folders:
                self._delete_folder(self.profile_folder_name, name)

        return sorted(self._removed_folders), sorted(self._removed_files), self._modified_files

    def save(self):
        for folder in self._removed_folders:
            shutil.rmtree(folder, ignore_errors=True)
        for file in self._removed_files:
            try:
                os.remove(file)
            except FileNotFoundError:
                pass
        for path, data in self._modified_files.items():
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, 'w') as f:
                f.write(data)
        self._removed_folders = set()
        self._removed_files = set()
        self._modified_files = {}

    def _generate_by_command_group(self, profile_folder_name, command_group):
        assert command_group.command_groups or command_group.commands

        command_group_folder_names = self._command_group_folder_names(*command_group.names)
        cur_folders, cur_files = self._list_package(profile_folder_name, *command_group_folder_names)

        folders = set()
        if command_group.command_groups:
            for sub_group in command_group.command_groups.values():
                assert sub_group.names[:-1] == command_group.names, f"Invalid command group name: {sub_group.names}"
                self._generate_by_command_group(profile_folder_name=profile_folder_name, command_group=sub_group)
                sub_group_folder_name = sub_group.names[-1].replace('-', '_')
                folders.add(sub_group_folder_name)

        # delete other folders
        del_folders = cur_folders.difference(folders)
        for name in del_folders:
            self._delete_folder(profile_folder_name, *command_group_folder_names, name)

        files = set()
        if command_group.commands:
            for command in command_group.commands.values():
                assert command.names[:-1] == command_group.names, f"Invalid command name: {command.names}"
                cmd_file_name = self._command_file_name(command.names[-1])
                if cmd_file_name in cur_files:
                    if command.cfg:
                        # configuration attached, that means to update command file
                        self._generate_by_command(profile_folder_name, command)
                else:
                    assert command.cfg is not None
                    self._generate_by_command(profile_folder_name, command)
                files.add(cmd_file_name)
            if command_group.wait_command:
                command = command_group.wait_command
                cmd_file_name = self._command_file_name(command.names[-1])
                if command.cfg:
                    # configuration attached, that means to update command file
                    self._generate_by_command(profile_folder_name, command, is_wait=True)
                files.add(cmd_file_name)

        # update __cmd_group.py file
        file_name = '__cmd_group.py'
        tmpl = get_templates()['aaz']['group'][file_name]
        data = tmpl.render(
            node=command_group
        )
        self._update_file(profile_folder_name, *command_group_folder_names, file_name, data=data)
        files.add(file_name)

        # update __init__.py file
        file_name = '__init__.py'
        tmpl = get_templates()['aaz']['group'][file_name]
        data = tmpl.render(
            file_names=sorted(files)
        )
        self._update_file(profile_folder_name, *command_group_folder_names, file_name, data=data)
        files.add(file_name)

        # delete other files
        del_files = cur_files.difference(files)
        for name in del_files:
            self._delete_file(profile_folder_name, *command_group_folder_names, name)

    def _generate_by_command(self, profile_folder_name, command, is_wait=False):
        assert isinstance(command.cfg, CMDCommand)
        file_name = self._command_file_name(command.names[-1])
        tmpl = get_templates()['aaz']['command']['_cmd.py']
        try:
            data = tmpl.render(
                leaf=AzCommandGenerator(command, is_wait=is_wait)
            )
        except exceptions.InvalidAPIUsage as err:
            err.message = f"CommandGenerationError: {' '.join(command.names)}: {err.message}"
            raise err
        self._update_file(
            profile_folder_name, *self._command_group_folder_names(*command.names[:-1]), file_name, data=data)

    # folder operations
    def _get_path(self, *names):
        return os.path.join(self.aaz_folder, *names)

    def _delete_folder(self, *names):
        path = self._get_path(*names)
        if os.path.exists(path):
            assert os.path.isdir(path), f'Invalid folder path {path}'
            self._removed_folders.add(path)

    def _delete_file(self, *names):
        path = self._get_path(*names)
        if os.path.exists(path):
            assert os.path.isfile(path), f'Invalid file path {path}'
            self._removed_files.add(path)

    def _update_file(self, *names, data):
        path = self._get_path(*names)
        if os.path.exists(path):
            assert os.path.isfile(path), f'Invalid file path {path}'
        self._modified_files[path] = data

    def _exist_file(self, *names):
        path = self._get_path(*names)
        if os.path.exists(path):
            assert os.path.isfile(path), f'Invalid file path {path}'
            return True
        return False

    def _list_package(self, *names):
        path = self._get_path(*names)
        folder_names = []
        file_names = []
        if os.path.exists(path):
            assert os.path.isdir(path), f'Invalid folder path {path}'
            for name in os.listdir(path):
                sub_path = os.path.join(path, name)
                if os.path.isfile(sub_path):
                    if name.endswith('.py'):
                        file_names.append(name)
                elif os.path.isdir(sub_path):
                    init_file = os.path.join(sub_path, '__init__.py')
                    if os.path.exists(init_file) and os.path.isfile(init_file):
                        folder_names.append(name)
        return set(folder_names), set(file_names)

    @staticmethod
    def _command_file_name(name):
        return f"_{to_snake_case(name)}.py"

    @staticmethod
    def _command_group_folder_names(*names):
        return [name.replace('-', '_') for name in names]



import os
import webbrowser

import click
from flask.cli import pass_script_info, show_server_banner, DispatchingApp, SeparatedPathType
from flask.helpers import get_debug_flag, get_env
from utils.config import Config


def is_port_in_use(host, port):
    import socket
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((host, port)) == 0


@click.command("run", short_help="Run a development server.")
@click.option(
    "--host", "-h",
    default=Config.HOST,
    required=not Config.HOST,
    callback=Config.validate_and_setup_host,
    help="The interface to bind to."
)
@click.option(
    "--port", "-p",
    default=Config.PORT,
    required=not Config.PORT,
    callback=Config.validate_and_setup_port,
    help="The port to bind to."
)
@click.option(
    "--aaz-path", '-a',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.AAZ_PATH,
    required=not Config.AAZ_PATH,
    callback=Config.validate_and_setup_aaz_path,
    expose_value=False,
    help="The local path of aaz repo."
)
@click.option(
    "--swagger-path", '-s',
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
    default=Config.SWAGGER_PATH,
    callback=Config.validate_and_setup_swagger_path,
    expose_value=False,
    help="The local path of azure-rest-api-specs repo. Official repo is https://github.com/Azure/azure-rest-api-specs"
)
@click.option(
    "--swagger-module-path", "--sm",
    type=click.Path(file_okay=False, dir_okay=True, readable=True, resolve_path=True),
    default=Config.SWAGGER_MODULE_PATH,
    callback=Config.validate_and_setup_swagger_module_path,
    expose_value=False,
    help="The local path of swagger in module level. It can be substituted for --swagger-path."
)
@click.option(
    "--module", '-m',
    default=Config.DEFAULT_SWAGGER_MODULE,
    callback=Config.validate_and_setup_default_swagger_module,
    expose_value=False,
    help="The default swagger module. It is required when using --swagger-module-path."
)
@click.option(
    "--resource-provider", "--rp",
    default=Config.DEFAULT_RESOURCE_PROVIDER,
    callback=Config.validate_and_setup_default_resource_provider,
    expose_value=False,
    help="The default swagger resource provider."
)
@click.option(
    "--cli-path", '-c',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.CLI_PATH,
    required=not Config.CLI_PATH,
    callback=Config.validate_and_setup_cli_path,
    expose_value=False,
    help="The local path of azure-cli repo. Official repo is https://github.com/Azure/azure-cli"
)
@click.option(
    "--cli-extension-path", '-e',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.CLI_EXTENSION_PATH,
    required=not Config.CLI_EXTENSION_PATH,
    callback=Config.validate_and_setup_cli_extension_path,
    expose_value=False,
    help="The local path of azure-cli-extension repo. Official repo is https://github.com/Azure/azure-cli-extensions"
)
@click.option(
    "--workspaces-path", '-w',
    type=click.Path(file_okay=False, dir_okay=True, writable=True, readable=True, resolve_path=True),
    default=Config.AAZ_DEV_WORKSPACE_FOLDER,
    required=not Config.AAZ_DEV_WORKSPACE_FOLDER,
    callback=Config.validate_and_setup_aaz_dev_workspace_folder,
    expose_value=False,
    help="The folder to load and save workspaces."
)
@click.option(
    "--reload/--no-reload",
    default=None,
    help="Enable or disable the reloader. By default the reloader "
         "is active if debug is enabled.",
)
@click.option(
    "--debugger/--no-debugger",
    default=None,
    help="Enable or disable the debugger. By default the debugger "
         "is active if debug is enabled.",
)
@click.option(
    "--eager-loading/--lazy-loading",
    default=None,
    help="Enable or disable eager loading. By default eager "
         "loading is enabled if the reloader is disabled.",
)
@click.option(
    "--with-threads/--without-threads",
    default=True,
    help="Enable or disable multithreading.",
)
@click.option(
    "--extra-files",
    default=None,
    type=SeparatedPathType(),
    help=(
            "Extra files that trigger a reload on change. Multiple paths"
            f" are separated by {os.path.pathsep!r}."
    ),
)
@click.option(
    "--quiet", '-q',
    is_flag=True,
    default=False,
    is_eager=True,
    help="Without open web browser page."
)
@pass_script_info
def run_command(
        info, host, port, reload, debugger, eager_loading, with_threads, extra_files, quiet
):
    """Run a local development server.

    This server is for development purposes only. It does not provide
    the stability, security, or performance of production WSGI servers.

    The reloader and debugger are enabled by default if
    FLASK_ENV=development or FLASK_DEBUG=1.
    """
    debug = get_debug_flag()

    if reload is None:
        reload = debug

    if debugger is None:
        debugger = debug

    show_server_banner(get_env(), debug, info.app_import_path, eager_loading)
    app = DispatchingApp(info.load_app, use_eager_loading=eager_loading)

    if is_port_in_use(host, port):
        raise ValueError(f"The port '{port}' already been used in '{host}', please specify a new port in '--port' argument.")

    from werkzeug.serving import run_simple

    if quiet:
        print(f'Please open http://{host}:{port}/')
    else:
        webbrowser.open_new(f'http://{host}:{port}/')

    run_simple(
        host,
        port,
        app,
        use_reloader=reload,
        use_debugger=debugger,
        threaded=with_threads,
        extra_files=extra_files,
    )



import os
from packaging import version
from .plane import PlaneEnum


class Config:

    MIN_CLI_CORE_VERSION = version.parse("2.51.0")

    AAZ_PATH = os.environ.get("AAZ_PATH", None)
    SWAGGER_PATH = os.environ.get("AAZ_SWAGGER_PATH", None)
    SWAGGER_MODULE_PATH = os.environ.get("AAZ_SWAGGER_MODULE_PATH", None)

    DEFAULT_PLANE = PlaneEnum.Mgmt
    DEFAULT_SWAGGER_MODULE = os.environ.get("AAZ_SWAGGER_MODULE", None)  # use '/' to join sub modules
    DEFAULT_RESOURCE_PROVIDER = os.environ.get("AAZ_SWAGGER_RESOURCE_PROVIDER", None)

    CLI_PATH = os.environ.get("AAZ_CLI_PATH", None)
    CLI_EXTENSION_PATH = os.environ.get("AAZ_CLI_EXTENSION_PATH", None)

    AAZ_DEV_FOLDER = os.path.expanduser(
        os.environ.get("AAZ_DEV_FOLDER", os.path.join("~", ".aaz_dev"))
    )
    AAZ_DEV_WORKSPACE_FOLDER = os.path.expanduser(
        os.environ.get("AAZ_DEV_WORKSPACE_FOLDER", os.path.join(AAZ_DEV_FOLDER, "workspaces"))
    )

    # Flask configurations
    HOST = os.environ.get("AAZ_HOST", '127.0.0.1')
    PORT = int(os.environ.get("AAZ_PORT", 5000))
    STATIC_FOLDER = '../ui'
    STATIC_URL_PATH = '/'

    CLI_DEFAULT_PROFILE = 'latest'

    CLI_PROFILES = [
        CLI_DEFAULT_PROFILE,
        "2020-09-01-hybrid",
        "2019-03-01-hybrid",
        "2018-03-01-hybrid",
        "2017-03-09-profile",
    ]

    @classmethod
    def validate_and_setup_host(cls, ctx, param, value):
        cls.HOST = value
        return cls.HOST

    @classmethod
    def validate_and_setup_port(cls, ctx, param, value):
        cls.PORT = value
        return cls.PORT

    @classmethod
    def validate_and_setup_aaz_path(cls, ctx, param, value):
        # TODO: verify folder structure
        if value != cls.AAZ_PATH:
            cls.AAZ_PATH = os.path.expanduser(value)
            if not os.path.exists(cls.AAZ_PATH):
                raise ValueError(f"Path '{cls.AAZ_PATH}' does not exist.")
        return cls.AAZ_PATH

    @classmethod
    def validate_and_setup_swagger_path(cls, ctx, param, value):
        # TODO: verify folder structure
        if value != cls.SWAGGER_PATH:
            cls.SWAGGER_PATH = os.path.expanduser(value)
            if not os.path.exists(cls.SWAGGER_PATH):
                raise ValueError(f"Path '{cls.SWAGGER_PATH}' does not exist.")
        return cls.SWAGGER_PATH

    @classmethod
    def validate_and_setup_swagger_module_path(cls, ctx, param, value):
        # TODO: verify folder structure
        if value != cls.SWAGGER_MODULE_PATH:
            cls.SWAGGER_MODULE_PATH = os.path.expanduser(value)
            if not os.path.exists(cls.SWAGGER_MODULE_PATH):
                raise ValueError(f"Path '{cls.SWAGGER_MODULE_PATH}' does not exist.")
        return cls.SWAGGER_MODULE_PATH

    @classmethod
    def validate_and_setup_default_swagger_module(cls, ctx, param, value):
        cls.DEFAULT_SWAGGER_MODULE = value
        return cls.DEFAULT_SWAGGER_MODULE

    @classmethod
    def validate_and_setup_default_resource_provider(cls, ctx, param, value):
        cls.DEFAULT_RESOURCE_PROVIDER = value
        return cls.DEFAULT_RESOURCE_PROVIDER

    @classmethod
    def validate_and_setup_cli_path(cls, ctx, param, value):
        # TODO: verify folder structure
        if value != cls.CLI_PATH:
            cls.CLI_PATH = os.path.expanduser(value)
            if not os.path.exists(cls.CLI_PATH):
                raise ValueError(f"Path '{cls.CLI_PATH}' does not exist.")
        return cls.CLI_PATH

    @classmethod
    def validate_and_setup_cli_extension_path(cls, ctx, param, value):
        # TODO: verify folder structure
        if value != cls.CLI_EXTENSION_PATH:
            cls.CLI_EXTENSION_PATH = os.path.expanduser(value)
            if not os.path.exists(cls.CLI_EXTENSION_PATH):
                raise ValueError(f"Path '{cls.CLI_EXTENSION_PATH}' does not exist.")
        return cls.CLI_EXTENSION_PATH

    @classmethod
    def validate_and_setup_aaz_dev_workspace_folder(cls, ctx, param, value):
        # TODO: verify folder
        if value != cls.AAZ_DEV_WORKSPACE_FOLDER:
            cls.AAZ_DEV_WORKSPACE_FOLDER = os.path.expanduser(value)
            if os.path.exists(cls.AAZ_DEV_WORKSPACE_FOLDER) and not os.path.isdir(cls.AAZ_DEV_WORKSPACE_FOLDER):
                raise ValueError(f"Path '{cls.AAZ_DEV_WORKSPACE_FOLDER}' is not a folder.")
        return cls.AAZ_DEV_WORKSPACE_FOLDER


__all__ = ["Config"]



from schematics.models import Model
from schematics.types import URLType, BooleanType, StringType

from .fields import XmsTextField


class XML(Model):
    """
    A metadata object that allows for more fine-tuned XML model definitions.
    When using arrays, XML element names are not inferred (for singular/plural forms) and the name property should be used to add that information. See examples for expected behavior.
    """

    name = StringType()  # Replaces the name of the element/attribute used for the described schema property. When defined within the Items Object (items), it will affect the name of the individual XML elements within the list. When defined alongside type being array (outside the items), it will affect the wrapping element and only if wrapped is true. If wrapped is false, it will be ignored.
    namespace = URLType()  # The URL of the namespace definition.
    prefix = StringType()  # The prefix to be used for the name.
    attribute = BooleanType(default=False)  # Declares whether the property definition translates to an attribute instead of an element. Default value is false.
    wrapped = BooleanType(default=False)  # MAY be used only for an array definition. Signifies whether the array is wrapped (for example, <books><book/><book/></books>) or unwrapped (<book/><book/>). Default value is false. The definition takes effect only when defined alongside type being array (outside the items).

    x_ms_text = XmsTextField()



from command.model.configuration._output import CMDObjectOutput, CMDArrayOutput, CMDStringOutput


class AzOutputGenerator:

    def __init__(self, output, cmd_ctx):
        self._output = output
        self._cmd_ctx = cmd_ctx

    @property
    def ref(self):
        if isinstance(self._output, (CMDObjectOutput, CMDArrayOutput, CMDStringOutput)) and self._output.ref:
            ref, is_selector = self._cmd_ctx.get_variant(self._output.ref)
            if is_selector:
                return f'{ref}.required()'
            else:
                return ref
        return None

    @property
    def next_link(self):
        if isinstance(self._output, (CMDArrayOutput, )) and self._output.next_link:
            next_link, is_selector = self._cmd_ctx.get_variant(self._output.next_link)
            assert not is_selector
            return next_link
        return None

    @property
    def client_flatten(self):
        if isinstance(self._output, (CMDObjectOutput, CMDArrayOutput)):
            return self._output.client_flatten is True
        elif isinstance(self._output, CMDStringOutput):
            return False
        else:
            raise NotImplementedError()

    @property
    def value(self):
        if isinstance(self._output, CMDStringOutput):
            return self._output.value
        else:
            raise NotImplementedError()



from command.model.configuration import CMDJsonSubresourceSelector, CMDArrayIndexBase, CMDObjectIndexBase, \
    CMDObjectIndexDiscriminator, CMDObjectIndexAdditionalProperties, CMDObjectIndex, CMDArrayIndex, CMDSchema, \
    CMDObjectIndexAdditionalProperties, CMDSelectorIndex
from utils.case import to_camel_case


class AzJsonSelectorGenerator:

    def __init__(self, cmd_ctx, selector):
        assert isinstance(selector, CMDJsonSubresourceSelector)
        self._cmd_ctx = cmd_ctx
        self._selector = selector
        self._cmd_ctx.set_selector(selector)
        self.name = self._cmd_ctx.get_variant(selector.var, name_only=True)
        self.variant_key, is_selector_variant = cmd_ctx.get_variant(selector.ref)
        assert not is_selector_variant

        self._json = self._selector.json

    @property
    def cls_name(self):
        return f'{to_camel_case(self.name)}Selector'

    def iter_scopes_for_get(self):
        for scope in self._iter_scopes():
            yield scope

    def iter_scopes_for_set(self):
        for scope in self._iter_scopes(is_set=True):
            yield scope

    def _iter_scopes(self, is_set=False):
        scope_name = 'result'
        scope_define = self.variant_key
        for scope in _iter_selector_scopes_by_index_base(self._json, scope_name, scope_define, [], None, self._cmd_ctx, is_set):
            yield scope


def _iter_selector_scopes_by_index(index, scope_name, scope_define, idx_lines, cmd_ctx, is_set):
    assert isinstance(index, (CMDObjectIndex, CMDArrayIndex))
    if scope_define is not None:
        yield scope_name, scope_define, idx_lines, None, None, False
    scope_define = f"{scope_name}.{index.name}"
    for scope in _iter_selector_scopes_by_index_base(index, scope_name, scope_define, [], None, cmd_ctx, is_set):
        yield scope


def _iter_selector_scopes_by_index_base(index, scope_name, scope_define, idx_lines, previous_identifiers, cmd_ctx, is_set):

    def _handle_idx_lines_for_end():
        if previous_identifiers and is_set:
            for identifier in previous_identifiers:
                assert isinstance(identifier, CMDSchema)
                assert identifier.arg
                if identifier.name == "[Index]" or identifier.name.startswith('[]'):
                    assert idx_lines == ["idx = next(filters)[0]"]
                    assert scope_define == f"{scope_name}[idx]"
                    idx_lines[0] = f"idx = next(filters, [len({scope_name})])[0]"
                    if identifier.name == "[Index]":
                        arg_keys, hide = cmd_ctx.get_argument(identifier.arg)
                        assert not hide
                        idx_lines.append(f"{arg_keys} = idx")
                    return
                elif identifier.name == "{Key}":
                    assert idx_lines == ["idx = next(filters)[0]"]
                    assert scope_define == f"{scope_name}[idx]"
                    arg_keys, hide = cmd_ctx.get_argument(identifier.arg)
                    assert not hide
                    idx_lines[0] = f"idx = next(filters, [{arg_keys}.to_serialized_data()])[0]"
                    return

    if isinstance(index, CMDObjectIndexBase):
        is_end = False
        if index.prop:
            assert isinstance(index.prop, CMDSelectorIndex)
            yield scope_name, scope_define, idx_lines, None, None, is_end
            for scope in _iter_selector_scopes_by_index(index.prop, scope_name, None, [], cmd_ctx, is_set):
                yield scope
        elif index.discriminator:
            yield scope_name, scope_define, idx_lines, None, None, is_end
            next_scope_define = f"[{scope_name}]"
            for scope in _iter_selector_scopes_by_index_base(
                    index.discriminator, scope_name, next_scope_define, [], None, cmd_ctx, is_set):
                yield scope
        elif index.additional_props:
            assert isinstance(index.additional_props, CMDObjectIndexAdditionalProperties)
            if index.additional_props.item:
                assert index.additional_props.identifiers
                filter_builder = f"{scope_name}.items()"
                filters = []
                for identifier in index.additional_props.identifiers:
                    assert isinstance(identifier, CMDSchema)
                    assert identifier.arg
                    arg_keys, hide = cmd_ctx.get_argument(identifier.arg)
                    assert not hide
                    if identifier.name == '{Key}':
                        filter_key = "[0]"
                        filter_value = arg_keys
                        filter_is_constant = False
                    else:
                        assert identifier.name.startswith('{}')
                        filter_key = f"[1]{identifier.name[2:]}"
                        filter_value = arg_keys
                        filter_is_constant = False
                    filters.append((filter_key, filter_value, filter_is_constant))

                yield scope_name, scope_define, idx_lines, filter_builder, filters, is_end
                next_idx_lines = [
                    "idx = next(filters)[0]"
                ]
                next_scope_define = f"{scope_name}[idx]"
                for scope in _iter_selector_scopes_by_index_base(
                        index.additional_props.item, scope_name, next_scope_define, next_idx_lines,
                        index.additional_props.identifiers, cmd_ctx, is_set):
                    yield scope

            else:
                assert not index.additional_props.identifiers
                is_end = True
                _handle_idx_lines_for_end()
                yield scope_name, scope_define, idx_lines, None, None, is_end
        else:
            is_end = True
            _handle_idx_lines_for_end()
            yield scope_name, scope_define, idx_lines, None, None, is_end
    elif isinstance(index, CMDObjectIndexDiscriminator):
        is_end = False
        assert scope_define == f"[{scope_name}]"
        filter_builder = f"enumerate({scope_name})"
        filter_key = f"[1].{index.property}"
        filter_value = index.value
        filter_is_constant = True
        filters = [(filter_key, filter_value, filter_is_constant)]
        if index.prop:
            yield scope_name, scope_define, idx_lines, filter_builder, filters, is_end
            next_idx_lines = [
                "idx = next(filters)[0]"
            ]
            next_scope_define = f"{scope_name}[idx]"
            for scope in _iter_selector_scopes_by_index(index.prop, scope_name, next_scope_define, next_idx_lines, cmd_ctx, is_set):
                yield scope
        elif index.discriminator:
            yield scope_name, scope_define, idx_lines, filter_builder, filters, is_end
            next_scope_define = f"[{scope_name}]"
            for scope in _iter_selector_scopes_by_index_base(
                    index.discriminator, scope_name, next_scope_define, [], None, cmd_ctx, is_set):
                yield scope
        else:
            # Not support to end in a discriminator
            raise NotImplementedError()
    elif isinstance(index, CMDArrayIndexBase):
        is_end = False
        if index.item:
            assert index.identifiers
            filter_builder = f"enumerate({scope_name})"
            filters = []
            for identifier in index.identifiers:
                assert isinstance(identifier, CMDSchema)
                assert identifier.arg
                arg_keys, hide = cmd_ctx.get_argument(identifier.arg)
                assert not hide
                if identifier.name == '[Index]':
                    filter_key = "[0]"
                    filter_value = arg_keys
                    filter_is_constant = False
                else:
                    assert identifier.name.startswith('[]')
                    filter_key = f"[1]{identifier.name[2:]}"
                    filter_value = arg_keys
                    filter_is_constant = False
                filters.append((filter_key, filter_value, filter_is_constant))

            yield scope_name, scope_define, idx_lines, filter_builder, filters, is_end

            next_idx_lines = [
                "idx = next(filters)[0]"
            ]
            next_scope_define = f"{scope_name}[idx]"
            for scope in _iter_selector_scopes_by_index_base(index.item, scope_name, next_scope_define, next_idx_lines, index.identifiers, cmd_ctx, is_set):
                yield scope
        else:
            assert not index.identifiers
            is_end = True
            _handle_idx_lines_for_end()
            yield scope_name, scope_define, idx_lines, None, None, is_end



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from schematics.models import Model
from schematics.types import ModelType

from ._content import CMDRequestJson
from ._fields import CMDVariantField


class CMDInstanceCreateAction(Model):
    POLYMORPHIC_KEY = None

    # properties as tags
    ref = CMDVariantField(required=True, deserialize_from=["ref", "instance"])

    @classmethod
    def _claim_polymorphic(cls, data):
        if cls.POLYMORPHIC_KEY is None:
            return False

        if isinstance(data, dict):
            return cls.POLYMORPHIC_KEY in data
        elif isinstance(data, CMDInstanceCreateAction):
            return hasattr(data, cls.POLYMORPHIC_KEY)
        return False

    def generate_args(self, ref_args):
        raise NotImplementedError()

    def reformat(self, **kwargs):
        raise NotImplementedError()

    def register_cls(self, **kwargs):
        return NotImplementedError()


# json instance create
class CMDJsonInstanceCreateAction(CMDInstanceCreateAction):
    POLYMORPHIC_KEY = "json"

    # properties as nodes
    json = ModelType(CMDRequestJson, required=True)

    def generate_args(self, ref_args):
        return self.json.generate_args(ref_args=ref_args, is_update_action=False)

    def reformat(self, **kwargs):
        self.json.reformat(**kwargs)

    def register_cls(self, **kwargs):
        self.json.register_cls(**kwargs)



from ._resource import Resource
from ._resource_provider import ResourceProvider
from ._swagger_module import SwaggerModule, DataPlaneModule, MgmtPlaneModule
from ._swagger_specs import SwaggerSpecs, SingleModuleSwaggerSpecs
from ._swagger_loader import SwaggerLoader



from cli.model.common._fields import CLIProfileNameField
from schematics.models import Model
from schematics.types import ModelType, DictType

from ._command_group import CLIAtomicCommandGroup


class CLIAtomicProfile(Model):
    name = CLIProfileNameField(required=True)
    command_groups = DictType(
        field=ModelType(CLIAtomicCommandGroup),
        serialized_name="commandGroups",
        deserialize_from="commandGroups"
    )

    class Options:
        serialize_when_none = False

    @property
    def profile_folder_name(self):
        profile_folder_name = self.name.lower().replace('-', '_')
        if profile_folder_name != "latest":
            # for rest profiles such as 2019-03-01-hybrid, the folder name starts with digit,
            # it's not a valid python package name.
            profile_folder_name = "profile_" + profile_folder_name
        return profile_folder_name



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from schematics.models import Model
from schematics.types import ModelType, DictType, StringType

from ._profile import CLIViewProfile


class CLIModule(Model):
    name = StringType(required=True)
    folder = StringType()
    profiles = DictType(
        field=ModelType(CLIViewProfile),
    )

    class Options:
        serialize_when_none = False



from schematics.models import Model
from schematics.types import StringType, PolyModelType, ListType

from ._arg import CMDArg, CMDClsArgBase, CMDObjectArgBase, CMDArrayArgBase
from utils import exceptions


class CMDArgGroup(Model):
    # properties as tags
    name = StringType(required=True)

    # properties as nodes
    args = ListType(PolyModelType(CMDArg, allow_subclasses=True), min_size=1)

    def reformat(self, **kwargs):
        for arg in self.args:
            try:
                arg.reformat(**kwargs)
            except exceptions.VerificationError as err:
                err.payload['details'] = {
                    "type": "Argument",
                    "options": arg.options,
                    "var": arg.var,
                    "details": err.payload['details']
                }
                raise err
        self.args = sorted(self.args, key=lambda a: a.var)

    def register_cls(self, cls_register_map):
        if self.args:
            for arg in self.args:
                _iter_over_arg_for_cls_register(arg, cls_register_map)


def _iter_over_arg_for_cls_register(arg, cls_register_map):
    if arg is None:
        return

    if isinstance(arg, CMDClsArgBase):
        cls_name = arg.type[1:]
        if cls_name not in cls_register_map:
            cls_register_map[cls_name] = {
                "implement": None,
                "refers": []
            }
        cls_register_map[cls_name]['refers'].append(arg)
        return

    if getattr(arg, 'cls', None):
        cls_name = arg.cls
        if cls_name not in cls_register_map:
            cls_register_map[cls_name] = {
                "implement": None,
                "refers": []
            }
        cls_register_map[cls_name]['implement'] = arg

    if isinstance(arg, CMDObjectArgBase):
        if arg.args:
            for sub_arg in arg.args:
                _iter_over_arg_for_cls_register(sub_arg, cls_register_map)

        if arg.additional_props and arg.additional_props.item:
            _iter_over_arg_for_cls_register(arg.additional_props.item, cls_register_map)

    elif isinstance(arg, CMDArrayArgBase):
        _iter_over_arg_for_cls_register(arg.item, cls_register_map)



import os

from utils.plane import PlaneEnum
from ._swagger_module import MgmtPlaneModule, DataPlaneModule
from utils.exceptions import ResourceNotFind, InvalidAPIUsage


class SwaggerSpecs:

    def __init__(self, folder_path):
        self._folder_path = folder_path

    @property
    def _spec_folder_path(self):
        return os.path.join(self._folder_path, 'specification')

    def get_mgmt_plane_modules(self, plane):
        modules = []
        for name in os.listdir(self._spec_folder_path):
            module = self.get_mgmt_plane_module(name, plane=plane)
            if module:
                modules.append(module)
        return modules

    def get_mgmt_plane_module(self, *names, plane):
        if not names:
            return None
        name = names[0]
        if not PlaneEnum.is_valid_swagger_module(plane=plane, module_name=name):
            return None

        path = os.path.join(self._spec_folder_path, name, 'resource-manager')
        if not os.path.isdir(path):
            return None
        module = MgmtPlaneModule(plane=plane, name=name, folder_path=path)
        for name in names[1:]:
            path = os.path.join(path, name)
            if not os.path.isdir(path):
                return None
            module = MgmtPlaneModule(plane=plane, name=name, folder_path=path, parent=module)
        return module

    def get_data_plane_modules(self, plane):
        modules = []
        for name in os.listdir(self._spec_folder_path):
            module = self.get_data_plane_module(name, plane=plane)
            if module:
                modules.append(module)
        return modules

    def get_data_plane_module(self, *names, plane):
        if not names:
            return None
        name = names[0]
        if not PlaneEnum.is_valid_swagger_module(plane=plane, module_name=name):
            return None

        path = os.path.join(self._spec_folder_path, name, 'data-plane')
        if not os.path.isdir(path):
            return None
        module = DataPlaneModule(plane=plane, name=name, folder_path=path)
        for name in names[1:]:
            path = os.path.join(path, name)
            if not os.path.isdir(path):
                return None
            module = DataPlaneModule(plane=plane, name=name, folder_path=path, parent=module)
        return module


class SingleModuleSwaggerSpecs:

    def __init__(self, folder_path, module_name):
        if not os.path.isdir(folder_path):
            raise ValueError(f"Path not exist: {folder_path}")
        self._folder_path = folder_path
        self._module_name = module_name

    def get_mgmt_plane_modules(self, plane):
        names = self._module_name.split('/')
        if not PlaneEnum.is_valid_swagger_module(plane=plane, module_name=names[0]):
            raise InvalidAPIUsage(f"{names[0]} is not a valid mgmt plane module")

        path = os.path.join(self._folder_path, 'resource-manager')
        if not os.path.isdir(path):
            raise ResourceNotFind(
                f"Cannot find manage plane module '{self._module_name}'", payload=f"{path} is not exist")

        module = None
        for name in names:
            module = MgmtPlaneModule(plane=plane, name=name, folder_path=None, parent=module)
        module.folder_path = path
        assert module is not None
        return [module]

    def get_mgmt_plane_module(self, *names, plane):
        if not names:
            return None
        name = names[0]
        if not PlaneEnum.is_valid_swagger_module(plane=plane, module_name=name):
            return None

        module_str = '/'.join([plane, *names])
        module = None
        for m in self.get_mgmt_plane_modules(plane):
            if str(m) == module_str:
                module = m
                break
        return module

    def get_data_plane_modules(self, plane):
        names = self._module_name.split('/')
        if not PlaneEnum.is_valid_swagger_module(plane=plane, module_name=names[0]):
            raise InvalidAPIUsage(f"{names[0]} is not a supported data plane module.")

        path = os.path.join(self._folder_path, 'data-plane')
        if not os.path.isdir(path):
            raise ResourceNotFind(
                f"Cannot find data plane module '{self._module_name}'", payload=f"{path} is not exist")

        module = None
        for name in names:
            module = DataPlaneModule(plane=plane, name=name, folder_path=None, parent=module)
        module.folder_path = path
        assert module is not None
        return [module]

    def get_data_plane_module(self, *names, plane):
        if not names:
            return None
        name = names[0]
        if not PlaneEnum.is_valid_swagger_module(plane=plane, module_name=name):
            return None

        module_str = '/'.join([plane, *names])
        module = None
        for m in self.get_mgmt_plane_modules(plane):
            if str(m) == module_str:
                module = m
                break
        return module



from schematics.models import Model
from schematics.types import StringType, URLType


class ExternalDocumentation(Model):
    """Allows referencing an external resource for extended documentation."""

    description = StringType()  # A short description of the target documentation. GFM syntax can be used for rich text representation.
    url = URLType(required=True)  # The URL for the target documentation. Value MUST be in the format of a URL.
